from math import pi, cos
from numpy import empty
from .latticesection import LatticeSection
from .latticepanel import LatticePanel
from .latticestrip import LatticeStrip
from pygeom.geom3d import ihat, Point, Coordinate

class LatticeSheet(object):
    sect1 = None
    sect2 = None
    bspace = None
    yspace = None
    mirror = None
    levec = None
    cord = None
    nrml = None
    strps = None
    pnls = None
    mind = None
    def __init__(self, sect1: LatticeSection, sect2: LatticeSection):
        self.sect1 = sect1
        self.sect2 = sect2
        self.update()
    def update(self):
        if self.sect1.mirror or self.sect2.mirror:
            self.mirror = True
        else:
            self.mirror = False
        self.bspace, self.yspace = inherit_spacing(self)
        self.levec = self.sect2.pnt-self.sect1.pnt
        vecz = (ihat**self.levec).to_unit()
        vecy = (vecz**ihat).to_unit()
        self.cord = Coordinate(self.sect1.pnt, ihat, vecy, vecz)
        self.strps = []
        self.pnls = []
    def mesh_strips(self, lsid: int):
        self.strps = []
        pnta = self.sect1.pnt
        vecr = self.levec
        crda = self.sect1.chord
        crdb = self.sect2.chord
        crdr = crdb-crda
        anga = self.sect1.angle
        angb = self.sect2.angle
        angr = angb-anga
        lenb = len(self.bspace)
        for i in range(lenb-1):
            bd1 = self.bspace[i]
            bd2 = self.bspace[i+1]
            yd = self.yspace[i]
            bspc = (yd-bd1)/(bd2-bd1)
            pnt1 = pnta+bd1*vecr
            pnt2 = pnta+bd2*vecr
            crd1 = crda+bd1*crdr
            crd2 = crda+bd2*crdr
            strp = LatticeStrip(lsid, pnt1, pnt2, crd1, crd2, bspc=bspc)
            ang1 = anga+bd1*angr
            ang2 = anga+bd2*angr
            strp.set_angles(ang1, ang2)
            strp.sht = self
            self.strps.append(strp)
            lsid += 1
        return lsid
    def add_panel(self, pnl: LatticePanel):
        pnl.sht = self
        self.pnls.append(pnl)
    def set_control_points_and_normals(self):
        for pnl in self.pnls:
            pntc = pnl.return_panel_point()
            # pnl.set_coordinate(self.cord)
            pnl.set_control_point(pntc)
            pnl.set_camber_angle(0.0)
    def __repr__(self):
        return '<LatticeSheet>'

def inherit_spacing(sht: LatticeSheet):
    if sht.mirror:
        if sht.sect2.bspace is None:
            bspace = [0.0, 1.0]
        else:
            bspace = [1.0-bd for bd in sht.sect2.bspace]
            bspace.reverse()
        if sht.sect2.yspace is None:
            yspace = [0.5]
        else:
            yspace = [1.0-yd for yd in sht.sect2.yspace]
            yspace.reverse()
    else:
        if sht.sect1.bspace is None:
            bspace = [0.0, 1.0]
        else:
            bspace = sht.sect1.bspace
        if sht.sect1.yspace is None:
            yspace = [0.5]
        else:
            yspace = sht.sect1.yspace
    # if sht.sect1.bspace is not None:
    #     if sht.sect1.mirror:
    #         if sht.sect2.bspace is not None:
    #             bspace = [1.0-bd for bd in sht.sect2.bspace]
    #             bspace.reverse()
    #         else:
    #             bspace = [0.0, 1.0]
    #     else:
    #         bspace = sht.sect1.bspace
    # elif sht.sect2.bspace is not None:
    #     bspace = [1.0-bd for bd in sht.sect2.bspace]
    #     bspace.reverse()
    # else:
    #     bspace = [0.0, 1.0]
    # if sht.sect1.yspace is not None:
    #     if sht.sect1.mirror:
    #         if sht.sect2.yspace is None:
    #             yspace = [1.0-yd for yd in sht.sect2.yspace]
    #             yspace.reverse()
    #         else:
    #             yspace = [0.5]
    #     else:
    #         yspace = sht.sect1.yspace
    # elif sht.sect2.yspace is not None:
    #     yspace = [1.0-yd for yd in sht.sect2.yspace]
    #     yspace.reverse()
    # else:
    #     yspace = [0.5]
    return bspace, yspace



from math import pi, atan2, degrees
from pygeom.geom3d import Point, Vector, ihat
from .latticepanel import LatticePanel

class LatticeStrip(object):
    lsid = None
    pnt1 = None
    pnt2 = None
    crd1 = None
    crd2 = None
    ang1 = None
    ang2 = None
    bspc = None
    leni = None
    pnti = None
    pnls = None
    msid = None
    sht = None
    lent = None
    dyt = None
    dzt = None
    dst = None
    _crd = None
    _ang = None
    _area = None
    _avecrd = None
    def __init__(self, lsid: int, pnt1: Point, pnt2: Point, crd1: float, crd2: float, bspc: float=0.5):
        self.lsid = lsid
        self.pnt1 = pnt1
        self.pnt2 = pnt2
        self.crd1 = crd1
        self.crd2 = crd2
        self.bspc = bspc
        self.update()
    def update(self):
        self.pnls = []
        self.leni = self.pnt2-self.pnt1
        self.pnti = self.pnt1+self.bspc*self.leni
        self.lent = Vector(0.0, self.leni.y, self.leni.z)
        self.dyt = self.leni.y
        self.dzt = self.leni.z
        self.dst = (self.dyt**2+self.dzt**2)**0.5
    def set_angles(self, ang1: float, ang2: float):
        self.ang1 = ang1
        self.ang2 = ang2
    def add_panel(self, pnl):
        pnl.strp = self
        self.pnls.append(pnl)
        self.sht.add_panel(pnl)
    def set_mirror(self, mstrp):
        self.msid = mstrp.lsid
        for i in range(len(self.pnls)):
            mpnl = mstrp.pnls[i]
            pnl = self.pnls[i]
            pnl.mpid = mpnl.lpid
    def return_points(self, percrd: float):
        pnt1 = self.pnt1+self.crd1*percrd*ihat
        pnt2 = self.pnt2+self.crd2*percrd*ihat
        return pnt1, pnt2
    def trefftz_velocity(self, pnt: Point):
        r = Point(0.0, pnt.y, pnt.z)
        ra = Point(0.0, self.pnt1.y, self.pnt1.z)
        rb = Point(0.0, self.pnt2.y, self.pnt2.z)
        a = r-ra
        b = r-rb
        am = a.return_magnitude()
        bm = b.return_magnitude()
        x = ihat
        axx = a**x
        bxx = b**x
        vel = (axx/am**2-bxx/bm**2)/2/pi
        return vel
    @property
    def nrmt(self):
        return self.sht.cord.dirz
    @property
    def dihedral(self):
        return degrees(atan2(self.dzt, self.dyt))
    @property
    def chord(self):
        if self._crd is None:
            self._crd = self.crd1+(self.crd2-self.crd1)*self.bspc
        return self._crd
    @property
    def avechord(self):
        if self._avecrd is None:
            self._avecrd = (self.crd1+self.crd2)/2
        return self._avecrd
    @property
    def angle(self):
        if self._ang is None:
            self._ang = self.ang1+(self.ang2-self.ang1)*self.bspc
        return self._ang
    @property
    def area(self):
        if self._area is None:
            self._area = self.dst*(self.crd1+self.crd2)/2
        return self._area
    def __repr__(self):
        return '<LatticeStrip {:d}>'.format(self.lsid)
    def __str__(self):
        frmstr = 'LatticeStrip\t{:d}\t{:.5f}\t{:.5f}\t{:}'
        return frmstr.format(self.lsid, self.pnt1, self.pnt2, self.msid)
    def __format__(self, format_spec: str):
        frmstr = 'LatticeStrip\t{:d}\t{:'
        frmstr += format_spec
        frmstr += '}\t{:'
        frmstr += format_spec
        frmstr += '}\t{:}'
        return frmstr.format(self.lsid, self.pnt1, self.pnt2, self.msid)



from .latticesystem import LatticeSystem, latticesystem_from_json
from .latticesurface import LatticeSurface
from .latticesection import LatticeSection
from .latticesheet import LatticeSheet
from .latticestrip import LatticeStrip
from .latticepanel import LatticePanel
from .latticeresult import LatticeResult
from .latticeoptimum import LatticeOptimum

cfrm = '.5f'
dfrm = '.7f'
efrm = '.4f'



class NACA4(object):
    code = None
    _mt = None
    _mc = None
    _pc = None
    def __init__(self, code: str):
        self.code = code
    @property
    def mt(self):
        if self._mt is None:
            self._mt = float(self.code[2])/10+float(self.code[3])/100
        return self._mt
    @property
    def mc(self):
        if self._mc is None:
            self._mc = float(self.code[0])/100
        return self._mc
    @property
    def pc(self):
        if self._pc is None:
            self._pc = float(self.code[1])/10
        return self._pc
    def return_camber(self, xc: float):
        if xc < self.pc:
            yc = self.mc/self.pc**2*(2*self.pc*xc-xc**2)
        else:
            yc = self.mc/(1-self.pc)**2*((1-2*self.pc)+2*self.pc*xc-xc**2)
        return yc
    def return_camber_slope(self, xc: float):
        if xc < self.pc:
            dydx = self.mc/self.pc**2*(2*self.pc-2*xc)
        else:
            dydx = self.mc/(1-self.pc)**2*(2*self.pc-2*xc)
        return dydx
    def return_camber_angle(self, xc: float):
        from math import degrees, atan
        dydx = self.return_camber_slope(xc)
        return degrees(atan(dydx))



from math import pi, radians, cos, sin, atan2
from pygeom.geom3d import Point, Vector, Coordinate, ihat

class LatticePanel(object):
    lpid = None
    pnts = None
    cspc = None
    bspc = None
    pntc = None
    pnti = None
    pnta = None
    pntb = None
    leni = None
    alc = None
    mpid = None
    strp = None
    sht = None
    crd = None
    area = None
    def __init__(self, lpid: int, pnts: list, cspc: float=0.25, bspc: float=0.5):
        self.lpid = lpid
        self.pnts = pnts
        self.cspc = cspc
        self.bspc = bspc
        self.update()
    def update(self):
        pnt1 = self.pnts[0]
        pnt2 = self.pnts[1]
        pnt3 = self.pnts[2]
        pnt4 = self.pnts[3]
        veca = pnt3-pnt1
        vecb = pnt4-pnt2
        self.pnta = pnt1+self.cspc*veca
        self.pntb = pnt2+self.cspc*vecb
        self.leni = self.pntb-self.pnta
        self.pnti = self.pnta+self.bspc*self.leni
        self.th = atan2(self.leni.z, self.leni.y)
        crda = veca.return_magnitude()
        crdb = vecb.return_magnitude()
        self.crd = crda+(crdb-crda)*0.5
        self.area = self.leni.return_magnitude()*(veca.return_magnitude()+vecb.return_magnitude())/2
    def return_panel_point(self):
        pnt1 = self.pnts[0]
        pnt2 = self.pnts[1]
        pnt3 = self.pnts[2]
        pnt4 = self.pnts[3]
        veca = pnt3-pnt1
        vecb = pnt4-pnt2
        pnta = pnt1+0.75*veca
        pntb = pnt2+0.75*vecb
        lenc = pntb-pnta
        pntc = pnta+self.bspc*lenc
        return pntc
    def set_control_point(self, pntc: Point):
        self.pntc = pntc
    def set_camber_angle(self, alpha: float):
        self.alpha = alpha
    @property
    def nrml(self):
        als = radians(self.strp.angle)
        alc = radians(self.alpha)
        th = self.th
        return Vector(-sin(alc - als), -sin(th)*cos(alc - als), cos(th)*cos(alc - als))
    @property
    def dnda(self):
        als = radians(self.strp.angle)
        alc = radians(self.alpha)
        th = self.th
        return Vector(cos(alc - als), -sin(th)*sin(alc - als), sin(alc - als)*cos(th))
    def velocity(self, pnt: Point):
        r = pnt
        ra = self.pnta
        rb = self.pntb
        a = r-ra
        b = r-rb
        am = a.return_magnitude()
        bm = b.return_magnitude()
        x = ihat
        vel = Vector(0.0, 0.0, 0.0)
        if pnt != self.pnti:
            axb = a**b
            if axb.return_magnitude() != 0.0:
                den = am*bm+a*b
                vel += axb/den*(1/am+1/bm)
        axx = a**x
        if axx.return_magnitude() != 0.0:
            den = am-a*x
            vel += axx/den/am
        bxx = b**x
        if bxx.return_magnitude() != 0.0:
            den = bm-b*x
            vel -= bxx/den/bm
        vel = vel/(4*pi)
        return vel
    def __repr__(self):
        return '<LatticePanel {:d}>'.format(self.lpid)
    def __str__(self):
        frmstr = 'LatticePanel\t{:d}\t{:.5f}\t{:.5f}\t{:.5f}\t{:.5f}\t{:}'
        pnt1 = self.pnts[0]
        pnt2 = self.pnts[1]
        pnt3 = self.pnts[2]
        pnt4 = self.pnts[3]
        return frmstr.format(self.lpid, pnt1, pnt2, pnt3, pnt4, self.mpid)
    def __format__(self, format_spec: str):
        frmstr = 'LatticePanel\t{:d}\t{:'
        frmstr += format_spec
        frmstr += '}\t{:'
        frmstr += format_spec
        frmstr += '}\t{:'
        frmstr += format_spec
        frmstr += '}\t{:'
        frmstr += format_spec
        frmstr += '}\t{:}'
        pnt1 = self.pnts[0]
        pnt2 = self.pnts[1]
        pnt3 = self.pnts[2]
        pnt4 = self.pnts[3]
        return frmstr.format(self.lpid, pnt1, pnt2, pnt3, pnt4, self.mpid)



from math import pi, cos, asin

def normalise_spacing(spacing: list) -> list:
    s = sorted(spacing)
    smin = min(s)
    smax = max(s)
    news = [(sval-smin)/(smax-smin) for sval in s]
    return news

def semi_elliptical_spacing(num: int) -> list:
    s = [cos(asin(1-float(i)/num)) for i in range(num+1)]
    s[0] = 0.0
    return s

def full_elliptical_spacing(num: int) -> list:
    th = [pi-asin(float(i)/num*2) if i < num/2 else asin(2*(num-float(i))/num) for i in range(num+1)]
    s = [(1+cos(thi))/2 for thi in th]
    return s

def semi_cosine_spacing(num: int) -> list:
    s = [cos(float(num-i)*pi/2/num) for i in range(num+1)]
    s[0] = 0.0
    return s

def full_cosine_spacing(num: int) -> list:
    s = [(1+cos(float(num-i)*pi/num))/2 for i in range(num+1)]
    return s

def equal_spacing(num: int) -> list:
    return [float(i)/num for i in range(num+1)]



from .camber import NACA4
from .spacing import normalise_spacing, equal_spacing
from .spacing import semi_cosine_spacing, semi_elliptical_spacing
from .spacing import full_cosine_spacing, full_elliptical_spacing
from .elliptical import Elliptical
from .bell import Bell
from math import pi

def elliptical_lift_distribution(y: list, b: float, L: float):
    return [4/pi/b*L*(1-(2*yi/b)**2)**0.5 for yi in y]

def bell_lift_distribution(y: list, b: float, L: float):
    return [16/3/pi/b*L*(1-(2*yi/b)**2)**1.5 for yi in y]



from pygeom.geom3d import Vector, jhat
from numpy.matlib import zeros, empty, matrix
from math import pi, radians, cos, sin
from matplotlib.pyplot import figure
from .latticesystem import LatticeSystem

class LatticeResult(object):
    name = None
    sys = None
    rho = None
    alpha = None
    beta = None
    speed = None
    _vfs = None
    _qfs = None
    _udc = None
    _ulc = None
    _uyc = None
    _gam = None
    _gmat = None
    _avv = None
    _afv = None
    _amv = None
    _phi = None
    _pmat = None
    _nfvel = None
    _nffrc = None
    _nfmom = None
    _nffrctot = None
    _nfmomtot = None
    _trlft = None
    _trdrg = None
    _trfrc = None
    _trwsh = None
    _strpy = None
    _Cx = None
    _Cy = None
    _Cz = None
    _Cl = None
    _Cm = None
    _Cn = None
    _CL = None
    _CDi = None
    _CY = None
    _CL_ff = None
    _CDi_ff = None
    _CY_ff = None
    _e = None
    def __init__(self, name: str, sys: LatticeSystem):
        self.name = name
        self.sys = sys
    def reset(self):
        for attr in self.__dict__:
            if attr[0] == '_':
                self.__dict__[attr] = None
    def set_conditions(self, alpha: float=0.0, beta: float=0.0, speed: float=1.0, rho: float=1.0):
        self.alpha = alpha
        self.beta = beta
        self.speed = speed
        self.rho = rho
    def set_freestream_velocity(self, vfs: Vector):
        self.vfs = vfs
        self.udc = self.vfs.to_unit()
        self.ulc = (self.udc**jhat).to_unit()
        self.uyc = (self.udc**self.ulc).to_unit()
        num = len(self.sys.pnls)
        self.avv = empty((num, 1), dtype=Vector)
        self.afv = empty((num, 1), dtype=Vector)
        self.amv = empty((num, 1), dtype=Vector)
        for i, pnl in enumerate(self.sys.pnls):
            self.avv[i, 0] = self.vfs
            self.afv[i, 0] = self.avv[i, 0]**pnl.leni
            self.amv[i, 0] = (pnl.pnti-self.sys.rref)**self.afv[i, 0]
    def set_gam(self, gam: list):
        self._gam = gam
        self._gmat = matrix(self._gam, dtype=float).transpose()
    def set_phi(self, phi: list):
        if len(phi) != len(self.sys.strps):
            raise Exception('The length of phi must equal the number of strips.')
        self._phi = phi
        self._pmat = matrix(self._phi, dtype=float).transpose()
    def set_lift_distribution(self, l: list, rho: float, speed: float):
        if len(l) != len(self.sys.strps):
            raise Exception('The length of l must equal the number of strips.')
        # nrml = [strpi.leni.y for strpi in self.sys.strps]
        self._phi = [li/rho/speed for li in l]
        self._pmat = matrix(self._phi, dtype=float).transpose()
    @property
    def gmat(self):
        if self._gmat is None:
            gamx = self.sys.gam[:, 0]
            gamy = self.sys.gam[:, 1]
            gamz = self.sys.gam[:, 2]
            self._gmat = gamx*self.vfs.x+gamy*self.vfs.y+gamz*self.vfs.z
        return self._gmat
    @property
    def gam(self):
        if self._gam is None:
            self._gam = [self.gmat[i, 0] for i in range(self.gmat.shape[0])]
        return self._gam
    @property
    def pmat(self):
        if self._pmat is None:
            num = len(self.sys.strps)
            self._pmat = zeros((num, 1))
            for i, strp in enumerate(self.sys.strps):
                for pnl in strp.pnls:
                    self._pmat[i, 0] += self.gam[pnl.lpid]
        return self._pmat
    @property
    def phi(self):
        if self._phi is None:
            self._phi = [self.pmat[i, 0] for i in range(self.pmat.shape[0])]
        return self._phi
    @property
    def vfs(self):
        if self._vfs is None:
            if self.alpha is None:
                self.alpha = 0.0
            if self.beta is None:
                self.beta = 0.0
            if self.speed is None:
                self.speed = 1.0
            alrad = radians(self.alpha)
            btrad = radians(self.beta)
            cosal = cos(alrad)
            sinal = sin(alrad)
            cosbt = cos(btrad)
            sinbt = sin(btrad)
            self._vfs = Vector(cosal*cosbt, -sinbt, sinal*cosbt)*self.speed
        return self._vfs
    @property
    def qfs(self):
        if self._qfs is None:
            self._qfs = self.rho*self.speed**2/2
        return self._qfs
    @property
    def udc(self):
        if self._udc is None:
            self._udc = self.vfs.to_unit()
        return self._udc
    @property
    def ulc(self):
        if self._ulc is None:
            self._ulc = (self.udc**jhat).to_unit()
        return self._ulc
    @property
    def uyc(self):
        if self._uyc is None:
            self._uyc = (self.ulc**self.udc).to_unit()
        return self._uyc
    @property
    def avv(self):
        if self._avv is None:
            num = len(self.sys.pnls)
            self._avv = empty((num, 1), dtype=Vector)
            for i in range(num):
                self._avv[i, 0] = self.vfs
        return self._avv
    @property
    def afv(self):
        if self._afv is None:
            num = len(self.sys.pnls)
            self._afv = empty((num, 1), dtype=Vector)
            for i, pnl in enumerate(self.sys.pnls):
                self._afv[i, 0] = self.avv[i, 0]**pnl.leni
        return self._afv
    @property
    def amv(self):
        if self._amv is None:
            num = len(self.sys.pnls)
            self._amv = empty((num, 1), dtype=Vector)
            for i, pnl in enumerate(self.sys.pnls):
                self._amv[i, 0] = (pnl.pnti-self.sys.rref)**self.afv[i, 0]
        return self._amv
    @property
    def nfvel(self):
        if self._nfvel is None:
            tmp = self.sys.avg*self.gmat+self.avv
            self._nfvel = [tmp[i, 0] for i in range(tmp.shape[0])]
        return self._nfvel
    @property
    def nffrc(self):
        if self._nffrc is None:
            rho = self.rho
            tmp = self.sys.afg*self.gmat+self.afv
            self._nffrc = [gam*tmp[i, 0]*rho for i, gam in enumerate(self.gam)]
        return self._nffrc
    @property
    def nfmom(self):
        if self._nfmom is None:
            rho = self.rho
            tmp = self.sys.amg*self.gmat+self.amv
            self._nfmom = [gam*tmp[i, 0]*rho for i, gam in enumerate(self.gam)]
        return self._nfmom
    @property
    def trlft(self):
        if self._trlft is None:
            rhoV = self.rho*self.speed
            self._trlft = [phi*self.sys.blg[i, 0]*rhoV for i, phi in enumerate(self.phi)]
        return self._trlft
    @property
    def trdrg(self):
        if self._trdrg is None:
            rho = self.rho
            tmp = self.sys.bdg*self.pmat
            self._trdrg = [phi*tmp[i, 0]*rho for i, phi in enumerate(self.phi)]
        return self._trdrg
    @property
    def trfrc(self):
        if self._trfrc is None:
            rhoV = self.rho*self.speed
            self._trfrc = [phi*self.sys.byg[i, 0]*rhoV for i, phi in enumerate(self.phi)]
        return self._trfrc
    @property
    def trwsh(self):
        if self._trwsh is None:
            tmp = self.sys.bvg*self.pmat
            self._trwsh = [tmp[i, 0] for i in range(tmp.shape[0])]
        return self._trwsh
    @property
    def nffrctot(self):
        if self._nffrctot is None:
            self._nffrctot = sum(self.nffrc)
        return self._nffrctot
    @property
    def nfmomtot(self):
        if self._nfmomtot is None:
            self._nfmomtot = sum(self.nfmom)
        return self._nfmomtot
    @property
    def Cx(self):
        if self._Cx is None:
            self._Cx = -self.nffrctot.x/self.qfs/self.sys.sref
        return self._Cx
    @property
    def Cy(self):
        if self._Cy is None:
            self._Cy = self.nffrctot.y/self.qfs/self.sys.sref
        return self._Cy
    @property
    def Cz(self):
        if self._Cz is None:
            self._Cz = -self.nffrctot.z/self.qfs/self.sys.sref
        return self._Cz
    @property
    def CL(self):
        if self._CL is None:
            self._CL = self.ulc*self.nffrctot/self.qfs/self.sys.sref
        return self._CL
    @property
    def CDi(self):
        if self._CDi is None:
            self._CDi = self.udc*self.nffrctot/self.qfs/self.sys.sref
        return self._CDi
    @property
    def CY(self):
        if self._CY is None:
            self._CY = self.uyc*self.nffrctot/self.qfs/self.sys.sref
        return self._CY
    @property
    def Cl(self):
        if self._Cl is None:
            self._Cl = -self.nfmomtot.x/self.qfs/self.sys.sref/self.sys.bref
        return self._Cl
    @property
    def Cm(self):
        if self._Cm is None:
            self._Cm = self.nfmomtot.y/self.qfs/self.sys.sref/self.sys.cref
        return self._Cm
    @property
    def Cn(self):
        if self._Cn is None:
            self._Cn = -self.nfmomtot.z/self.qfs/self.sys.sref/self.sys.bref
        return self._Cn
    @property
    def CL_ff(self):
        if self._CL_ff is None:
            self._CL_ff = sum(self.trlft)/self.qfs/self.sys.sref
        return self._CL_ff
    @property
    def CDi_ff(self):
        if self._CDi_ff is None:
            self._CDi_ff = sum(self.trdrg)/self.qfs/self.sys.sref
        return self._CDi_ff
    @property
    def CY_ff(self):
        if self._CY_ff is None:
            self._CY_ff = sum(self.trfrc)/self.qfs/self.sys.sref
        return self._CY_ff
    @property
    def e(self):
        if self._e is None:
            self._e = self.CL_ff**2/pi/self.sys.ar/self.CDi_ff
        return self._e
    def plot_panel_near_field_velocities(self, ax=None, component=None):
        if ax is None:
            fig = figure(figsize=(12, 8))
            ax = fig.gca()
            ax.grid(True)
        py = [pnl.pnti.y for pnl in self.sys.pnls]
        if component is None or component == 'x':
            vx = [vel.x for vel in self.nfvel]
            ax.plot(py, vx, label=self.name+' Velocity X')
        if component is None or component == 'y':
            vy = [vel.y for vel in self.nfvel]
            ax.plot(py, vy, label=self.name+' Velocity Y')
        if component is None or component == 'z':
            vz = [vel.z for vel in self.nfvel]
            ax.plot(py, vz, label=self.name+' Velocity Z')
        ax.legend()
        return ax
    def plot_phi_distribution(self, ax=None):
        if ax is None:
            fig = figure(figsize=(12, 8))
            ax = fig.gca()
            ax.grid(True)
        ax.plot(self.sys.strpy, self.phi, label=self.name)
        ax.legend()
        return ax
    def plot_trefftz_lift_distribution(self, ax=None):
        if ax is None:
            fig = figure(figsize=(12, 8))
            ax = fig.gca()
            ax.grid(True)
        trlft = [self.trlft[i]/strp.dyt for i, strp in enumerate(self.sys.strps)]
        ax.plot(self.sys.strpy, trlft, label=self.name)
        ax.legend()
        return ax
    def plot_trefftz_drag_distribution(self, ax=None):
        if ax is None:
            fig = figure(figsize=(12, 8))
            ax = fig.gca()
            ax.grid(True)
        trdrg = [self.trdrg[i]/strp.dst for i, strp in enumerate(self.sys.strps)]
        ax.plot(self.sys.strpy, trdrg, label=self.name)
        ax.legend()
        return ax
    def plot_trefftz_wash_distribution(self, ax=None):
        if ax is None:
            fig = figure(figsize=(12, 8))
            ax = fig.gca()
            ax.grid(True)
        ax.plot(self.sys.strpy, self.trwsh, label=self.name)
        ax.legend()
        return ax
    def plot_strip_wash_distribution(self, lsid: int, ax=None):
        if ax is None:
            fig = figure(figsize=(12, 8))
            ax = fig.gca()
            ax.grid(True)
        stwash = []
        for i in range(self.sys.bvg.shape[0]):
            stwash.append(self.sys.bvg[i, lsid])
        ax.plot(self.sys.strpy, stwash)
        return ax
    def print_near_field_total_loads(self):
        print(f'Total Force in X = {self.nffrctot.x}')
        print(f'Total Force in Y = {self.nffrctot.y}')
        print(f'Total Force in Z = {self.nffrctot.z}')
        print(f'Total Moment in X = {self.nfmomtot.x}')
        print(f'Total Moment in Y = {self.nfmomtot.y}')
        print(f'Total Moment in Z = {self.nfmomtot.z}')
    def print_aerodynamic_coefficients(self):
        from . import cfrm, dfrm, efrm
        if self.gam is not None:
            print('Cx = '+self.Cx.__format__(cfrm))
            print('Cy = '+self.Cy.__format__(cfrm))
            print('Cz = '+self.Cz.__format__(cfrm))
            print('Cl = '+self.Cl.__format__(cfrm))
            print('Cm = '+self.Cm.__format__(cfrm))
            print('Cn = '+self.Cn.__format__(cfrm))
            print('CL = '+self.CL.__format__(cfrm))
            print('CDi = '+self.CDi.__format__(dfrm))
            print('CY = '+self.CY.__format__(cfrm))
        if self.phi is not None:
            print('CL_ff = '+self.CL_ff.__format__(cfrm))
            print('CDi_ff = '+self.CDi_ff.__format__(dfrm))
            print('CY_ff = '+self.CY_ff.__format__(cfrm))
            print('e = '+self.e.__format__(efrm))
    def print_strip_forces(self):
        from py2md.classes import MDTable
        from math import atan
        table = MDTable()
        table.add_column('#', 'd')
        table.add_column('Yle', '.4f')
        table.add_column('Chord', '.4f')
        table.add_column('Area', '.4f')
        table.add_column('c cl', '.4f')
        table.add_column('ai', '.4f')
        table.add_column('cl_norm', '.4f')
        table.add_column('cl', '.4f')
        table.add_column('cd', '.4f')
        q = self.qfs
        for strp in self.sys.strps:
            j = strp.lsid
            yle = strp.pnti.y
            chord = strp.chord
            area = strp.area
            nrmfrc = Vector(0.0, 0.0, 0.0)
            for pnl in strp.pnls:
                nrmfrc += self.nffrc[pnl.lpid]/pnl.area*pnl.crd
            c_cl = nrmfrc.z/q
            ai = -self.trwsh[strp.lsid]
            cd = self.trdrg[strp.lsid]/q/area
            # cx = nrmfrc*self.udc/q/chord
            cy = nrmfrc*self.uyc/q/chord
            cz = nrmfrc*self.ulc/q/chord
            cf = Vector(0.0, cy, cz)
            cl_norm = cf*strp.nrmt
            cl = cl_norm
            # cd = nrmfrc*self.udc/q/chord
            # cd = cx
            table.add_row([j, yle, chord, area, c_cl, ai, cl_norm, cl, cd])
            # print(frmstr.format(j, yle, chord, area, c_cl, ai, cl_norm, cl, cd))
        print(table)
    def print_strip_coefficients(self):
        from py2md.classes import MDTable
        from math import atan
        table = MDTable()
        table.add_column('#', 'd')
        table.add_column('Chord', '.5f')
        table.add_column('Area', '.5f')
        table.add_column('cn', '.5f')
        table.add_column('ca', '.5f')
        table.add_column('cl', '.5f')
        table.add_column('cd', '.5f')
        table.add_column('dw', '.5f')
        table.add_column('cm le', '.5f')
        table.add_column('cm qc', '.5f')
        q = self.qfs
        for strp in self.sys.strps:
            j = strp.lsid
            chord = strp.chord
            area = strp.area
            force = Vector(0.0, 0.0, 0.0)
            momle = Vector(0.0, 0.0, 0.0)
            for pnl in strp.pnls:
                force += self.nffrc[pnl.lpid]
                momle += self.nfmom[pnl.lpid]
                rref = pnl.pnti-strp.pnti
                momle += rref**self.nffrc[pnl.lpid]
            cn = force*strp.nrmt/q/area
            ca = force.x/q/area
            cl = self.trlft[strp.lsid]/q/area
            cd = self.trdrg[strp.lsid]/q/area
            dw = -self.trwsh[strp.lsid]
            cmle = momle.y/q/area/chord
            rqc = Vector(-chord/4, 0.0, 0.0)
            momqc = momle+rqc**force
            cmqc = momqc.y/q/area/chord
            table.add_row([j, chord, area, cn, ca, cl, cd, dw, cmle, cmqc])
        print(table)
    def print_panel_forces(self):
        from py2md.classes import MDTable
        from math import atan
        table = MDTable()
        table.add_column('# P', 'd')
        table.add_column('# S', 'd')
        table.add_column('X', '.5f')
        table.add_column('Y', '.5f')
        table.add_column('Z', '.5f')
        table.add_column('DX', '.5f')
        table.add_column('Slope', '.5f')
        table.add_column('dCp', '.5f')
        q = self.qfs
        for pnl in self.sys.pnls:
            j = pnl.lpid
            k = pnl.strp.lsid
            x = pnl.pnti.x
            y = pnl.pnti.y
            z = pnl.pnti.z
            area = pnl.area
            frc = self.nffrc[pnl.lpid]
            nfrc = frc*pnl.nrml
            cp = nfrc/area/q
            chord = pnl.crd
            alc = pnl.alc
            alc = 0.0
            table.add_row([j, k, x, y, z, chord, alc, cp])
        print(table)
    def print_panel_near_field_results(self):
        from py2md.classes import MDTable
        from math import atan
        table = MDTable()
        table.add_column('# P', 'd')
        table.add_column('# S', 'd')
        table.add_column('Gamma', '.1f')
        table.add_column('Vx', '.1f')
        table.add_column('Vy', '.1f')
        table.add_column('Vz', '.1f')
        table.add_column('lx', '.4f')
        table.add_column('ly', '.4f')
        table.add_column('lz', '.4f')
        table.add_column('Fx', '.2f')
        table.add_column('Fy', '.2f')
        table.add_column('Fz', '.2f')
        for pnl in self.sys.pnls:
            j = pnl.lpid
            k = pnl.strp.lsid
            gam = self.gam[j]
            Vx = self.nfvel[j].x
            Vy = self.nfvel[j].y
            Vz = self.nfvel[j].z
            lx = pnl.leni.x
            ly = pnl.leni.y
            lz = pnl.leni.z
            Fx = self.nffrc[j].x
            Fy = self.nffrc[j].y
            Fz = self.nffrc[j].z
            table.add_row([j, k, gam, Vx, Vy, Vz, lx, ly, lz, Fx, Fy, Fz])
        print(table)
    def __str__(self):
        from py2md.classes import MDTable
        from . import cfrm, dfrm, efrm
        outstr = '# Lattice Result '+self.name+'\n'
        table = MDTable()
        table.add_column('Alpha (deg)', 'g', data=[self.alpha])
        table.add_column('Beta (deg)', 'g', data=[self.beta])
        table.add_column('Speed', 'g', data=[self.speed])
        table.add_column('Rho', 'g', data=[self.rho])
        outstr += table._repr_markdown_()
        if self.gam is not None:
            table = MDTable()
            table.add_column('Cx', cfrm, data=[self.Cx])
            table.add_column('Cy', cfrm, data=[self.Cy])
            table.add_column('Cz', cfrm, data=[self.Cz])
            table.add_column('Cl', cfrm, data=[self.Cl])
            table.add_column('Cm', cfrm, data=[self.Cm])
            table.add_column('Cn', cfrm, data=[self.Cn])
            outstr += table._repr_markdown_()
            table = MDTable()
            table.add_column('CL', cfrm, data=[self.CL])
            table.add_column('CDi', dfrm, data=[self.CDi])
            table.add_column('CY', cfrm, data=[self.CY])
            outstr += table._repr_markdown_()
        if self.phi is not None:
            table = MDTable()
            table.add_column('CL_ff', cfrm, data=[self.CL_ff])
            table.add_column('CDi_ff', dfrm, data=[self.CDi_ff])
            table.add_column('CY_ff', cfrm, data=[self.CY_ff])
            table.add_column('e', efrm, data=[self.e])
            outstr += table._repr_markdown_()
        return outstr
    def __repr__(self):
        return f'<LatticeResult: {self.name}>'
    def _repr_markdown_(self):
        return self.__str__()

class PanelResult(object):
    pnl = None
    gam = None
    vel = None
    frc = None
    mom = None
    def __init__(self, pnl):
        self.pnl = pnl
    def set_gam(self, gam: float):
        self.gam = gam
    def set_velocity(self, vel: Vector):
        self.vel = vel
    def set_force(self, frc: Vector):
        self.frc = frc
    def set_moment(self, mom: Vector):
        self.mom = mom

class StripResult(object):
    strp = None
    phi = None
    nfwsh = None
    nflft = None
    nfdrg = None
    nfmom = None
    pnlres = None
    def __init__(self, strp):
        self.strp = strp
        self.pnlres = []
    def add_panel_result(self, pnlres):
        self.pnlres.append(pnlres)



from math import pi, atan2
from numpy.matlib import zeros, empty
from numpy.linalg import solve, inv
from pygeom.geom3d import Point, Vector, ihat, jhat, zero_vector

class LatticeSystem(object):
    name = None
    srfcs = None
    strps = None
    mstrp = None
    pnls = None
    mpnl = None
    bref = None
    cref = None
    sref = None
    rref = None
    _gam = None
    _avg = None
    _afg = None
    _amg = None
    _avc = None
    _aic = None
    _afs = None
    _adc = None
    _bvg = None
    _bdg = None
    _blg = None
    _byg = None
    _bmg = None
    _ar = None
    _strpy = None
    def __init__(self, name: str, srfcs: list):
        self.name = name
        self.srfcs = srfcs
        self.mesh()
    def mesh(self):
        lsid = 0
        lpid = 0
        for srfc in self.srfcs:
            lsid, lpid = srfc.mesh(lsid, lpid)
        self.inherit()
    def inherit(self):
        strpdct = {}
        pnldct = {}
        istrp = 0
        ipnl = 0
        self.strps = []
        self.mstrp = []
        self.pnls = []
        self.mpnl = []
        for srfc in self.srfcs:
            for strp in srfc.strps:
                self.strps.append(strp)
                if strp.msid is None:
                    strpind = istrp
                    strpdct[strp.lsid] = istrp
                else:
                    strpind = strpdct[strp.msid]
                self.mstrp.append(strpind)
                istrp += 1
            pnls = srfc.return_panels()
            for pnl in pnls:
                self.pnls.append(pnl)
                if pnl.mpid is None:
                    pnlind = ipnl
                    pnldct[pnl.lpid] = ipnl
                else:
                    pnlind = pnldct[pnl.mpid]
                self.mpnl.append(pnlind)
                ipnl += 1
    def reset(self):
        for attr in self.__dict__:
            if attr[0] == '_':
                self.__dict__[attr] = None
    def set_reference_geometry(self, bref: float, cref: float, sref: float):
        self.bref = bref
        self.cref = cref
        self.sref = sref
        self._ar = None
    def set_reference_point(self, xref: float, yref: float, zref: float):
        self.rref = Point(xref, yref, zref)
    @property
    def gam(self):
        if self._gam is None:
            self._gam = -solve(self.aic, self.afs)
        return self._gam
    @property
    def avg(self):
        if self._avg is None:
            num = len(self.pnls)
            self._avg = empty((num, num), dtype=Vector)
            for i, pnli in enumerate(self.pnls):
                for j, pnlj in enumerate(self.pnls):
                    self._avg[i, j] = pnlj.velocity(pnli.pnti)
        return self._avg
    @property
    def afg(self):
        if self._afg is None:
            num = len(self.pnls)
            self._afg = empty((num, num), dtype=Vector)
            for i, pnl in enumerate(self.pnls):
                for j in range(num):
                    self._afg[i, j] = self.avg[i, j]**pnl.leni
        return self._afg
    @property
    def amg(self):
        if self._amg is None:
            num = len(self.pnls)
            self._amg = empty((num, num), dtype=Vector)
            for i, pnl in enumerate(self.pnls):
                for j in range(num):
                    self._amg[i, j] = (pnl.pnti-self.rref)**self.afg[i, j]
        return self._amg
    @property
    def avc(self):
        if self._avc is None:
            num = len(self.pnls)
            self._avc = empty((num, num), dtype=Vector)
            for i, pnli in enumerate(self.pnls):
                for j, pnlj in enumerate(self.pnls):
                    self._avc[i, j] = pnlj.velocity(pnli.pntc)
        return self._avc
    @property
    def aic(self):
        if self._aic is None:
            num = len(self.pnls)
            self._aic = zeros((num, num))
            for i, pnl in enumerate(self.pnls):
                for j in range(num):
                    self._aic[i, j] = self.avc[i, j]*pnl.nrml
        return self._aic
    @property
    def adc(self):
        if self._adc is None:
            num = len(self.pnls)
            self._adc = zeros((num, num))
            for i, pnl in enumerate(self.pnls):
                for j in range(num):
                    self._adc[i, j] = self.avc[i, j]*pnl.tang
        return self._adc
    @property
    def afs(self):
        if self._afs is None:
            num = len(self.pnls)
            self._afs = zeros((num, 3))
            for i, pnl in enumerate(self.pnls):
                self._afs[i, 0] = pnl.nrml.x
                self._afs[i, 1] = pnl.nrml.y
                self._afs[i, 2] = pnl.nrml.z
        return self._afs
    @property
    def bvg(self):
        if self._bvg is None:
            num = len(self.strps)
            self._bvg = zeros((num, num))
            for i, strpi in enumerate(self.strps):
                for j, strpj in enumerate(self.strps):
                    self._bvg[i, j] = strpj.trefftz_velocity(strpi.pnti)*strpi.nrmt
        return self._bvg
    @property
    def bdg(self):
        if self._bdg is None:
            num = len(self.strps)
            self._bdg = zeros((num, num))
            for i, strp in enumerate(self.strps):
                for j in range(num):
                    self._bdg[i, j] = -strp.dst*self.bvg[i, j]/2
        return self._bdg
    @property
    def blg(self):
        if self._blg is None:
            num = len(self.strps)
            self._blg = zeros((num, 1))
            for i, strp in enumerate(self.strps):
                self._blg[i, 0] = strp.lent.y
        return self._blg
    @property
    def byg(self):
        if self._byg is None:
            num = len(self.strps)
            self._byg = zeros((num, 1))
            for i, strp in enumerate(self.strps):
                self._byg[i, 0] = -strp.lent.z
        return self._byg
    @property
    def bmg(self):
        if self._bmg is None:
            num = len(self.strps)
            self._bmg = zeros((num, 1))
            for i, strp in enumerate(self.strps):
                self._bmg[i, 0] = strp.pnti.y*self.blg[i, 0]-strp.pnti.z*self.byg[i, 0]
        return self._bmg
    @property
    def strpy(self):
        if self._strpy is None:
            self._strpy = [strp.pnti.y for strp in self.strps]
        return self._strpy
    # def optimum_lift_distribution(self, Lspec: float, lspec: float=None):
        # from IPython.display import display
        # from pyhtml import HTMLMatrix
        # nump = len(self.strps)
        # numc = 1
        # if lspec is not None:
        #     numc += 2
        # amat = zeros((nump+numc, nump+numc))
        # bmat = zeros((nump+numc, 1))
        # amat[0:nump, 0:nump] = self.bdg + self.bdg.transpose()
        # amat[0:nump, nump] = self.blg
        # amat[nump, 0:nump] = self.blg.transpose()
        # bmat[nump, 0] = Lspec
        # bmg1 = self.bmg.copy()
        # bmg2 = self.bmg.copy()
        # hlfp = int(nump/2)
        # for i in range(hlfp):
        #     bmg1[i, 0] = 0.0
        # for i in range(hlfp, nump):
        #     bmg2[i, 0] = 0.0
        # if lspec is not None:
        #     amat[0:nump, nump+1] = bmg1
        #     amat[nump+1, 0:nump] = bmg1.transpose()
        #     bmat[nump+1, 0] = lspec
        #     amat[0:nump, nump+2] = bmg2
        #     amat[nump+2, 0:nump] = bmg2.transpose()
        #     bmat[nump+2, 0] = -1.0*lspec
        # xmat = solve(amat, bmat)
        # phi = [xmat[i, 0] for i in range(nump)]
        # lam = [xmat[nump+j, 0] for j in range(numc)]
        # Di = (xmat[0:nump, 0].transpose()*self.bdg*xmat[0:nump, 0])[0, 0]
        # L = (xmat[0:nump, 0].transpose()*self.blg)[0, 0]
        # l = (xmat[0:nump, 0].transpose()*bmg1)[0, 0]
        # amat = HTMLMatrix(amat)
        # display(amat)
        # return phi, lam, Di, L, l
    @property
    def ar(self):
        if self._ar is None:
            self._ar = self.bref**2/self.sref
        return self._ar
    def set_strip_alpha(self, alpha: list):
        for i, strp in enumerate(self.strps):
            strp._ang = alpha[i]
        self._aic = None
        self._afs = None
        self._gam = None
    def plot_surface_ipv(self):
        from ipyvolume import figure, plot_wireframe, show
        fig = figure()
        for srfc in self.srfcs:
            x, y, z = srfc.point_xyz()
            _ = plot_wireframe(x, y, z)
        show()
        return fig
    def plot_surface(self, view=None):
        from matplotlib.pyplot import figure
        from mpl_toolkits.mplot3d import axes3d
        # pltbox = PlotBox3D()
        fig = figure(figsize=(12,8))
        ax = fig.gca(projection='3d')
        ax.set_proj_type('ortho')
        ax.grid(False)
        # ax.set_aspect('equal')
        for srfc in self.srfcs:
            x, y, z = srfc.point_xyz()
            ax.plot_wireframe(x, y, z)
            # pltbox.update_box(x, y, z)
        if view == 'top':
            ax.view_init(90.0, 0.0)
        elif view == 'left':
            ax.view_init(0.0, 0.0)
        # pltbox.plot_box(ax)
        set_axes_equal(ax)
        fig.tight_layout()
    def print_strip_geometry(self):
        from py2md.classes import MDTable
        table = MDTable()
        table.add_column('#', 'd')
        table.add_column('Xle', '.5f')
        table.add_column('Yle', '.5f')
        table.add_column('Zle', '.5f')
        table.add_column('Chord', '.4f')
        table.add_column('Width', '.5f')
        table.add_column('Area', '.6f')
        table.add_column('Dihed', '.4f')
        table.add_column('Incid', '.4f')
        for strp in self.strps:
            j = strp.lsid
            xle = strp.pnti.x
            yle = strp.pnti.y
            zle = strp.pnti.z
            chord = strp.avechord
            width = strp.dst
            area = strp.area
            dihed = strp.dihedral
            angle = strp.angle
            table.add_row([j, xle, yle, zle, chord, width, area, dihed, angle])
        print(table)
    def print_panel_geometry(self):
        from py2md.classes import MDTable
        table = MDTable()
        table.add_column('#', 'd')
        table.add_column('X', '.5f')
        table.add_column('Y', '.5f')
        table.add_column('Z', '.5f')
        table.add_column('DX', '.5f')
        for pnl in self.pnls:
            j = pnl.lpid
            x = pnl.pnti.x
            y = pnl.pnti.y
            z = pnl.pnti.z
            dx = pnl.crd
            table.add_row([j, x, y, z, dx])
        print(table)
    def __repr__(self):
        return '<LatticeSystem: {:s}>'.format(self.name)
    def __str__(self):
        from py2md.classes import MDTable
        outstr = '# Lattice System '+self.name+'\n'
        table = MDTable()
        table.add_column('Name', 's', data=[self.name])
        table.add_column('Sref', 'g', data=[self.sref])
        table.add_column('cref', 'g', data=[self.cref])
        table.add_column('bref', 'g', data=[self.bref])
        table.add_column('xref', '.3f', data=[self.rref.x])
        table.add_column('yref', '.3f', data=[self.rref.y])
        table.add_column('zref', '.3f', data=[self.rref.z])
        outstr += table._repr_markdown_()
        table = MDTable()
        table.add_column('# Strips', 'd', data=[len(self.strps)])
        table.add_column('# Panels', 'd', data=[len(self.pnls)])
        outstr += table._repr_markdown_()
        return outstr
    def _repr_markdown_(self):
        return self.__str__()

def latticesystem_from_json(jsonfilepath: str):
    from .latticesurface import latticesurface_from_json
    from json import load
    with open(jsonfilepath, 'rt') as jsonfile:
        data = load(jsonfile)
    name = data['name']
    sfcs = []
    for surfdata in data['surfaces']:
        sfc = latticesurface_from_json(surfdata)
        sfcs.append(sfc)
    sys = LatticeSystem(name, sfcs)
    bref = data['bref']
    cref = data['cref']
    sref = data['sref']
    sys.set_reference_geometry(bref, cref, sref)
    xref = data['xref']
    yref = data['yref']
    zref = data['zref']
    sys.set_reference_point(xref, yref, zref)
    return sys

class PlotBox3D(object):
    xmin = None
    ymin = None
    zmin = None
    xmax = None
    ymax = None
    zmax = None
    def __init__(self):
        pass
    def update_box(self, x, y, z):
        if self.xmin is None:
            self.xmin = x.min()
        else:
            self.xmin = min(x.min(), self.xmin)
        if self.xmax is None:
            self.xmax = x.max()
        else:
            self.xmax = max(x.max(), self.xmax)
        if self.ymin is None:
            self.ymin = y.min()
        else:
            self.ymin = min(y.min(), self.ymin)
        if self.ymax is None:
            self.ymax = y.max()
        else:
            self.ymax = max(y.max(), self.ymax)
        if self.zmin is None:
            self.zmin = z.min()
        else:
            self.zmin = min(z.min(), self.zmin)
        if self.zmax is None:
            self.zmax = z.max()
        else:
            self.zmax = max(z.max(), self.zmax)
    def plot_box(self, ax):
        xrng = self.xmax-self.xmin
        yrng = self.ymax-self.ymin
        zrng = self.zmax-self.zmin
        xctr = (self.xmax+self.xmin)/2
        yctr = (self.ymax+self.ymin)/2
        zctr = (self.zmax+self.zmin)/2
        maxrng = max([xrng, yrng, zrng])
        xmin = xctr-maxrng/2
        ymin = yctr-maxrng/2
        zmin = zctr-maxrng/2
        xmax = xctr+maxrng/2
        ymax = yctr+maxrng/2
        zmax = zctr+maxrng/2
        # x = [xmin, xmin, xmin, xmin, xmax, xmax, xmax, xmax]
        # y = [ymin, ymin, ymax, ymax, ymin, ymin, ymax, ymax]
        # z = [zmin, zmax, zmin, zmax, zmin, zmax, zmin, zmax]
        # ax.scatter(x, y, z)
        ax.set_xlim3d(xmin, xmax)
        ax.set_ylim3d(ymin, ymax)
        ax.set_zlim3d(zmin, zmax)

def set_axes_radius(ax, origin, radius):
    ax.set_xlim3d([origin[0] - radius, origin[0] + radius])
    ax.set_ylim3d([origin[1] - radius, origin[1] + radius])
    ax.set_zlim3d([origin[2] - radius, origin[2] + radius])

def set_axes_equal(ax):
    from numpy import array, mean, max, abs
    '''Make axes of 3D plot have equal scale so that spheres appear as spheres,
    cubes as cubes, etc..  This is one possible solution to Matplotlib's
    ax.set_aspect('equal') and ax.axis('equal') not working for 3D.

    Input
      ax: a matplotlib axis, e.g., as output from plt.gca().
    '''

    limits = array([
        ax.get_xlim3d(),
        ax.get_ylim3d(),
        ax.get_zlim3d(),
    ])

    origin = mean(limits, axis=1)
    radius = 0.5 * max(abs(limits[:, 1] - limits[:, 0]))
    set_axes_radius(ax, origin, radius)



from numpy.matlib import zeros, matrix
from numpy.linalg import solve, norm
from .latticesystem import LatticeSystem
from .latticeresult import LatticeResult
from pygeom.geom3d import Point, Vector, zero_vector
from math import degrees
from matplotlib.pyplot import figure

class LatticeOptimum(object):
    name = None
    sys = None
    sym = None
    ind = None
    constr = None
    record = None
    rho = None
    speed = None
    res = None
    _phi = None
    _pmat = None
    _rhov = None
    _bdg = None
    _adg = None
    def __init__(self, name: str, sys: LatticeSystem, sym: bool=True):
        self.name = name
        self.sys = sys
        self.sym = sym
        self.update()
    def update(self):
        self.ind = []
        for strp in self.sys.strps:
            self.ind.append(None)
        if self.sym:
            ind = 0
            for strp in self.sys.strps:
                if strp.msid is not None:
                    self.ind[strp.lsid] = ind
                    self.ind[strp.msid] = ind
                    ind += 1
            for strp in self.sys.strps:
                if self.ind[strp.lsid] is None:
                    if strp.msid is None:
                        self.ind[strp.lsid] = ind
                        ind += 1
        else:
            ind = 0
            for strp in self.sys.strps:
                self.ind[strp.lsid] = ind
                ind += 1
    def set_conditions(self, speed: float=1.0, rho: float=1.0):
        self.speed = speed
        self.rho = rho
    def add_constraint(self, param: str, value: float, strplst: list=None, point: Point=None):
        if self.constr is None:
            self.constr = []
        constr = Constraint(self, param, value, point, strplst)
        self.constr.append(constr)
    def add_record(self, param: str, strplst: list=None, point: Point=None):
        if self.record is None:
            self.record = []
        record = Constraint(self, param, None, point, strplst)
        self.record.append(record)
    def initial_pmat(self):
        nump = max(self.ind)+1
        return zeros((nump, 1))
    @property
    def rhov(self):
        if self._rhov is None:
            self._rhov = self.rho*self.speed
        return self._rhov
    @property
    def bdg(self):
        if self._bdg is None:
            num = max(self.ind)+1
            self._bdg = zeros((num, num))
            for i, indi in enumerate(self.ind):
                for j, indj in enumerate(self.ind):
                    self._bdg[indi, indj] += self.rhov*self.sys.bdg[i, j]
        return self._bdg
    @property
    def adg(self):
        if self._adg is None:
            self._adg = self.bdg+self.bdg.transpose()
        return self._adg
    def old_iteration(self, pmat: matrix):
        nump = max(self.ind)+1
        num = nump+len(self.constr)
        amat = zeros((num, num))
        bmat = zeros((num, 1))
        amat[0:nump, 0:nump] = self.adg
        for i, constr in enumerate(self.constr):
            acv, ach = constr.return_matrices(pmat)
            amat[0:nump, nump+i] = acv
            amat[nump+i, 0:nump] = ach
            bmat[nump+i, 0] = constr.value
        xmat = solve(amat, bmat)
        pmat = xmat[0:nump, 0]
        lmat = xmat[nump:num, 0]
        return pmat, lmat
    def optimum_lift_distribution(self, crit=1e-12):
        pmat_cur = self.initial_pmat()
        pmat_new, lmat = self.old_iteration(pmat_cur)
        while norm(pmat_new-pmat_cur) > crit:
            pmat_cur = pmat_new
            pmat_new, lmat = self.old_iteration(pmat_cur)
        phi = [pmat_new[ind, 0] for ind in self.ind]
        lam = [lmat[i, 0] for i in range(len(lmat))]
        self._pmat = pmat_new
        self._lmat = lmat
        return phi, lam
    @property
    def phi(self):
        if self._phi is None:
            self._phi, _ = self.optimum_lift_distribution()
        return self._phi
    @property
    def pmat(self):
        if self._pmat is None:
            nump = max(self.ind)+1
            self._pmat = zeros((nump, 1))
            for i, ind in enumerate(self.ind):
                self._pmat[ind, 0] = self.phi[i]
        return self._pmat
    def set_phi(self, phi: list):
        if len(phi) != len(self.sys.strps):
            raise Exception('The length of phi must equal the number of strips.')
        self._phi = phi
    def set_lift_distribution(self, l: list, rho: float, speed: float):
        if len(l) != len(self.sys.strps):
            raise Exception('The length of l must equal the number of strips.')
        self._phi = [li/rho/speed for li in l]
        self.speed = speed
        self.rho = rho
    def return_induced_drag(self):
        return (self.pmat.transpose()*self.bdg*self.pmat)[0, 0]
    def print_report(self):
        print(f'{self.name} Report')
        Di = self.return_induced_drag()
        print(f'Di = {Di:.6f}')
        if self.constr is not None:
            for constr in self.constr:
                val = constr.evaluate(self.pmat)
                print(f'{constr.param} = {val:.6f}')
        if self.record is not None:
            for record in self.record:
                val = record.evaluate(self.pmat)
                print(f'{record.param} = {val:.6f}')
    def optimum_strip_twist_iteration(self):
        
        nump = len(self.sys.pnls)
        nums = len(self.sys.strps)

        popt = zeros((nums, 1))
        for i, phi in enumerate(self.phi):
            popt[i, 0] = phi

        dafsda = zeros((nump, nums))
        daicda = zeros((nump, nums))

        for i in range(nump):
            for j, strp in enumerate(self.sys.strps):
                for pnl in strp.pnls:
                    k = pnl.lpid
                    if k == i:
                        dafsda[i, j] += self.res.vfs*pnl.dnda
                        daicda[i, j] += self.sys.avc[i, k]*pnl.dnda

        dgda = -solve(self.sys.aic, dafsda+daicda*self.res.pmat)

        dpda = zeros((nums, nums))
        for i, strp in enumerate(self.sys.strps):
            for j in range(nums):
                for pnl in strp.pnls:
                    k = pnl.lpid
                    dpda[i, j] += dgda[k, j]

        dal = solve(dpda, popt-self.res.pmat)

        for i in range(nums):
            dal[i, 0] = degrees(dal[i, 0])

        da = dal.transpose().tolist()[0]

        al = [strp.angle for i, strp in enumerate(self.sys.strps)]
        alc = [al[i]+da[i] for i in range(nums)]

        self.sys.set_strip_alpha(alc)

        return dal
    def optimum_strip_twist(self, crit=1e-1):
        self.res = LatticeResult(self.name, self.sys)
        self.res.set_conditions(speed=self.speed, rho=self.rho)
        dal = self.optimum_strip_twist_iteration()
        nrmdal = norm(dal)
        iteration = 0
        self.res = LatticeResult(self.name, self.sys)
        self.res.set_conditions(speed=self.speed, rho=self.rho)
        dal = self.optimum_strip_twist_iteration()
        nrmdal = norm(dal)
        while nrmdal > crit:
            iteration += 1
            print(f'Iteration {iteration} - Convergence {nrmdal}')
            self.res = LatticeResult(self.name, self.sys)
            self.res.set_conditions(speed=self.speed, rho=self.rho)
            dal = self.optimum_strip_twist_iteration()
            nrmdal = norm(dal)
        self.res = LatticeResult(self.name, self.sys)
        self.res.set_conditions(speed=self.speed, rho=self.rho)
        al = [strp.angle for i, strp in enumerate(self.sys.strps)]
        return al
    def plot_strip_twist_distribution(self, ax=None):
        if ax is None:
            fig = figure(figsize=(12, 8))
            ax = fig.gca()
            ax.grid(True)
        y = [strp.pnti.y for i, strp in enumerate(self.sys.strps)]
        al = [strp.angle for i, strp in enumerate(self.sys.strps)]
        ax.plot(y, al, label=f'{self.name} Strip Twist')
        ax.legend()
        return ax
    def __repr__(self):
        return '<LatticeOptimum {:s}>'.format(self.name)
    def __str__(self):
        from py2md.classes import MDTable
        outstr = '# '+self.name+'\n'
        table = MDTable()
        table.add_column('Speed [m/s]', 'g', data=[self.speed])
        table.add_column('Rho [kg/m**3]', 'g', data=[self.rho])
        outstr += table._repr_markdown_()
        if self._phi is not None:
            table = MDTable()
            table.add_column('Label', 's')
            table.add_column('Type', 's')
            table.add_column('Value', 'g')
            table.add_row(['Di', 'Objective', self.return_induced_drag()])
            if self.constr is not None:
                for constr in self.constr:
                    val = constr.evaluate(self.pmat)
                    table.add_row([constr.param, 'Constraint', val])
            if self.record is not None:
                for record in self.record:
                    val = record.evaluate(self.pmat)
                    table.add_row([record.param, 'Record', val])
            outstr += table._repr_markdown_()
        return outstr
    def _repr_markdown_(self):
        return self.__str__()

class Constraint(object):
    opt = None
    param = None
    value = None
    point = None
    strplst = None
    _bcv = None
    _bcm = None
    def __init__(self, opt: LatticeOptimum, param: str, value: float, pnt: Point=None, strplst: list=None):
        self.opt = opt
        self.param = param
        self.value = value
        self.strplst = strplst
        self.update()
    def update(self):
        if self.strplst is None:
            self.strplst = []
            for strp in self.opt.sys.strps:
                self.strplst.append(strp.lsid)
        elif self.strplst == 'Mirrored':
            self.strplst = []
            for strp in self.opt.sys.strps:
                if strp.msid is not None:
                    self.strplst.append(strp.lsid)
        if self.point is None:
            self.point = self.opt.sys.rref
    @property
    def bcv(self):
        if self._bcv is None:
            num = max(self.opt.ind)+1
            if self.param == 'L':
                self._bcv = zeros((num, 1))
                for i in self.strplst:
                    ind = self.opt.ind[i]
                    self._bcv[ind, 0] += self.opt.rhov*self.opt.sys.blg[i, 0]
            elif self.param == 'Y':
                self._bcv = zeros((num, 1))
                for i in self.strplst:
                    ind = self.opt.ind[i]
                    self._bcv[ind, 0] += self.opt.rhov*self.opt.sys.bdg[i, 0]
            elif self.param == 'l':
                self._bcv = zeros((num, 1))
                for i in self.strplst:
                    strp = self.opt.sys.strps[i]
                    ind = self.opt.ind[i]
                    bli = self.opt.sys.blg[i, 0]
                    byi = self.opt.sys.byg[i, 0]
                    ryi = strp.pnti.y-self.point.y
                    rzi = strp.pnti.z-self.point.z
                    self._bcv[ind, 0] += self.opt.rhov*(ryi*bli-rzi*byi)
            elif self.param == 'm':
                self._bcv = zeros((num, 1))
                for i in self.strplst:
                    strp = self.opt.sys.strps[i]
                    ind = self.opt.ind[i]
                    bli = self.opt.sys.blg[i, 0]
                    rxi = strp.pnti.x-self.point.x
                    self._bcv[ind, 0] -= self.opt.rhov*rxi*bli
            elif self.param == 'n':
                self._bcv = zeros((num, 1))
                for i in self.strplst:
                    strp = self.opt.sys.strps[i]
                    ind = self.opt.ind[i]
                    byi = self.opt.sys.byg[i, 0]
                    rxi = strp.pnti.x-self.point.x
                    self._bcv[ind, 0] += self.opt.rhov*rxi*byi
        return self._bcv
    @property
    def bcm(self):
        if self._bcm is None:
            num = max(self.opt.ind)+1
            if self.param == 'm':
                self._bcm = zeros((num, num))
                for i in self.strplst:
                    strpi = self.opt.sys.strps[i]
                    indi = self.opt.ind[i]
                    rzi = strpi.pnti.z-self.point.z
                    for j, indj in enumerate(self.opt.ind):
                        bdi = self.opt.sys.bdg[i, j]
                        self._bcm[indi, indj] += self.opt.rhov*rzi*bdi
            elif self.param == 'n':
                self._bcm = zeros((num, num))
                for i in self.strplst:
                    strpi = self.opt.sys.strps[i]
                    indi = self.opt.ind[i]
                    ryi = strpi.pnti.y-self.point.y
                    for j, indj in enumerate(self.opt.ind):
                        bdi = self.opt.sys.bdg[i, j]
                        self._bcm[indi, indj] -= self.opt.rhov*ryi*bdi
        return self._bcm
    def return_matrices(self, pmat: matrix):
        ach = self.bcv.transpose()
        acv = self.bcv
        if self.bcm is not None:
            ach = ach + pmat.transpose()*self.bcm
            acv = acv + (self.bcm.transpose()+self.bcm)*pmat
        return acv, ach
    def evaluate(self, pmat: matrix):
        _, ach = self.return_matrices(pmat)
        value = (ach*pmat)[0, 0]
        return value



from math import asin, cos, pi
from pygeom.geom3d import Point, ihat
from pyvlm.tools.function import Function

class LatticeSection(object):
    pnt = None
    chord = None
    angle = None
    camber = None
    bspace = None
    yspace = None
    mirror = None
    def __init__(self, pnt: Point, chord: float, angle: float):
        self.pnt = pnt
        self.chord = chord
        self.angle = angle
        self.update()
    def update(self):
        self.mirror = False
        self.camber = Function([0.0, 1.0], [0.0, 0.0])
    def set_span_equal_spacing(self, numb: int):
        from pyvlm.tools import equal_spacing
        spc = equal_spacing(2*numb)
        self.bspace = spc[0::2]
        self.yspace = spc[1::2]
    def set_span_cosine_spacing(self, numb: int):
        from pyvlm.tools import full_cosine_spacing
        spc = full_cosine_spacing(2*numb)
        self.bspace = spc[0::2]
        self.yspace = spc[1::2]
    def set_span_semi_cosine_spacing(self, numb: int):
        from pyvlm.tools import semi_cosine_spacing
        spc = semi_cosine_spacing(2*numb)
        self.bspace = spc[0::2]
        self.yspace = spc[1::2]
    # def set_span_elliptical_spacing(self, numb: int):
    #     from pyvlm.tools import full_elliptical_spacing
    #     spc = full_elliptical_spacing(2*numb)
    #     self.bspace = spc[0::2]
    #     self.yspace = spc[1::2]
    # def set_span_semi_elliptical_spacing(self, numb: int):
    #     from pyvlm.tools import semi_elliptical_spacing
    #     spc = semi_elliptical_spacing(2*numb)
    #     self.bspace = spc[0::2]
    #     self.yspace = spc[1::2]
    def set_camber(self, xc: list, zc: list):
        xmin = min(xc)
        xmax = max(xc)
        xrng = xmax-xmin
        xnrm = [(xci-xmin)/xrng for xci in xc]
        znrm = [zci/xrng for zci in zc]
        self.camber = Function(xnrm, znrm)
    def return_mirror(self):
        pnt = Point(self.pnt.x, -self.pnt.y, self.pnt.z)
        chord = self.chord
        angle = self.angle
        sect = LatticeSection(pnt, chord, angle)
        sect.camber = self.camber
        sect.bspace = self.bspace
        sect.yspace = self.yspace
        sect.mirror = True
        return sect
    def return_point(self, percrd: float):
        return self.pnt+self.chord*percrd*ihat
    def get_camber(self, xc: float):
        return self.camber.cubic_interp(xc)
    def __repr__(self):
        return '<LatticeSection>'

def latticesecttion_from_json(sectdata: dict):
    xle = sectdata['xle']
    yle = sectdata['yle']
    zle = sectdata['zle']
    crd = sectdata['chord']
    if 'angle' in sectdata:
        ang = sectdata['angle']
    else:
        ang = 0.0
    pnt = Point(xle, yle, zle)
    sect = LatticeSection(pnt, crd, ang)
    if 'airfoil' in sectdata:
        sect.airfoil = sectdata['airfoil']
    if 'numb' in sectdata and 'bspace' in sectdata:
        numb = sectdata['numb']
        bspace = sectdata['bspace']
        if bspace == 'equal':
            sect.set_span_equal_spacing(numb)
        elif bspace == 'cosine':
            sect.set_span_cosine_spacing(numb)
        # elif bspace == 'elliptical':
        #     sect.set_span_elliptical_spacing(numb)
        elif bspace == 'semi-cosine':
            sect.set_span_semi_cosine_spacing(numb)
        # elif bspace == 'semi-elliptical':
        #     sect.set_span_semi_elliptical_spacing(numb)
    return sect



"""Module containing function for 1D interpolation"""

class Function(object):
    x = None
    y = None
    m = None
    dy2 = None
    def __init__(self, x, y):
        if isinstance(x, list) and isinstance(y, list):
            lenx = len(x)
            leny = len(y)
            if lenx == leny and lenx > 1 and leny > 1:
                ascending = True
                for i in range(lenx-1):
                    if x[i] > x[i+1]:
                        ascending = False
                if ascending:
                    self.x = x
                    self.y = y
        self.calc_d2y()
    def calc_m(self):
        num = len(self.x)
        self.m = []
        for i in range(num-1):
            dx = (self.x[i+1]-self.x[i])
            if dx == 0.:
                self.m.append(0.)
            else:
                dy = (self.y[i+1]-self.y[i])
                self.m.append(dy/dx)
    def linear_interp(self, xi):
        if self.m == None:
            self.calc_m()
        num = len(self.x)
        if isinstance(xi, list):
            numi = len(xi)
            yi = [None]*numi
            for i in range(numi):
                for j in range(num-1):
                    if xi[i] >= self.x[j] and xi[i] <= self.x[j+1]:
                        yi[i] = self.y[j]+self.m[j]*(xi[i]-self.x[j])
                        break
            return yi
        if isinstance(xi, int):
            xi = float(xi)
        if isinstance(xi, float):
            for j in range(num-1):
                if xi >= self.x[j] and xi <= self.x[j+1]:
                    yi = self.y[j]+self.m[j]*(xi-self.x[j])
                    break
            return yi
    def cumtrapz(self):
        A = [0.]
        for i in range(1,len(self.x)):
            A.append(A[i-1]+(self.y[i]+self.y[i-1])/2*(self.x[i]-self.x[i-1]))
        return A
    def calc_d2y(self):
        num = len(self.x)
        a = [0. for i in range(num)]
        b = [1. for i in range(num)]
        c = [0. for i in range(num)]
        r = [0. for i in range(num)]
        for i in range(1, num-1):
            dxA = self.x[i]-self.x[i-1]
            dxB = self.x[i+1]-self.x[i]
            dyA = self.y[i]-self.y[i-1]
            dyB = self.y[i+1]-self.y[i]
            a[i] = dxA/6
            b[i] = (dxA+dxB)/3
            c[i] = dxB/6
            r[i] = dyB/dxB-dyA/dxA
        gm = [0. for i in range(num)]
        bt = b[0]
        self.d2y = [0. for i in range(num)]
        self.d2y[0] = r[0]/bt
        for i in range(1, num):
            gm[i] = c[i-1]/bt
            bt = b[i]-a[i]*gm[i]
            self.d2y[i] = (r[i]-a[i]*self.d2y[i-1])/bt
        for i in range(num-2, 0, -1):
            self.d2y[i] -= gm[i+1]*self.d2y[i+1]
    def cubic_interp(self, xi):
        num = len(self.x)
        if isinstance(xi, list):
            numi = len(xi)
            yi = [None]*numi
            for i in range(numi):
                for j in range(num-1):
                    if xi[i] >= self.x[j] and xi[i] <= self.x[j+1]:
                        dx = (self.x[j+1]-self.x[j])
                        A = (self.x[j+1]-xi[i])/dx
                        B = (xi[i]-self.x[j])/dx
                        C = (A**3-A)*dx**2/6
                        D = (B**3-B)*dx**2/6
                        yi[i] = A*self.y[j]+B*self.y[j+1]+C*self.d2y[j]+D*self.d2y[j-1]
            return yi
        if isinstance(xi, int):
            xi = float(xi)
        if isinstance(xi, float):
            for j in range(num-1):
                if xi >= self.x[j] and xi <= self.x[j+1]:
                    dx = (self.x[j+1]-self.x[j])
                    A = (self.x[j+1]-xi)/dx
                    B = (xi-self.x[j])/dx
                    C = (A**3-A)*dx**2/6
                    D = (B**3-B)*dx**2/6
                    yi = A*self.y[j]+B*self.y[j+1]+C*self.d2y[j]+D*self.d2y[j-1]
            return yi
    def cubic_interp_eval(self, xi=None, level=0, sv1=0., sv2=0.):
        if xi == None:
            if level == 0:
                return self.y
        num = len(self.x)
        if xi == None:
            if level == -2:
                return self.d2y
        iy0 = [0. for i in range(num)]
        i2y0 = [0. for i in range(num)]
        iy0[0] = sv1
        i2y0[0] = sv2
        for i in range(1, num):
            xa = self.x[i-1]
            xb = self.x[i]
            ya = self.y[i-1]
            yb = self.y[i]
            ua = self.d2y[i-1]
            ub = self.d2y[i]
            iy0[i] = iy0[i-1]
            iy0[i] += (xb-xa)/2*ya
            iy0[i] += (xb-xa)/2*yb
            iy0[i] += (xa-xb)**3/24*ua
            iy0[i] += (xa-xb)**3/24*ub
            i2y0[i] = iy0[i-1]*(xb-xa)+i2y0[i-1]
            i2y0[i] += (2*xb**2+2*xa*xb-xa**2)/6*ya
            i2y0[i] += (xb**2-2*xa*xb-2*xa**2)/6*yb
            i2y0[i] += (7*xa**4-28*xa**3*xb+12*xa**2*xb**2+32*xa*xb**3-8*xb**4)/360*ua
            i2y0[i] += (8*xa**4-32*xa**3*xb-12*xa**2*xb**2+28*xa*xb**3-7*xb**4)/360*ub
        if xi == None:
            if level == 1:
                return iy0
            if level == 2:
                return i2y0
        return iy0, i2y0

def function_from_dict(dinp):
    if isinstance(dinp, dict):
        x = [k for k in sorted(dinp)]
        y = [dinp[k] for k in sorted(dinp)]
        return Function(x, y)



from math import pi, cos
from .latticestrip import LatticeStrip
from .latticesheet import LatticeSheet
from .latticepanel import LatticePanel

class LatticeSurface(object):
    name = None
    sects = None
    cspace = None
    xspace = None
    bdsit = None
    strps = None
    pnts = None
    pnls = None
    msht = None
    msect = None
    mirrored = None
    def __init__(self, name: str, sects: list, mirror: bool):
        self.name = name
        self.sects = sects
        self.mirror = mirror
        self.update()
    def update(self):
        if self.mirror and self.sects[0].pnt.y == 0.0:
            numsect = len(self.sects)
            newsects = []
            for i in range(numsect-1):
                sect = self.sects[numsect-1-i]
                msect = sect.return_mirror()
                newsects.append(msect)
            for sect in self.sects:
                newsects.append(sect)
            self.sects = newsects
        elif self.mirror and self.sects[0].pnt.y != 0.0:
            print(f'Warning: Cannot mirror {self.name}.')
            self.mirror = False
    def set_chord_distribution(self, cspace: list):
        from pyvlm.tools import normalise_spacing
        self.cspace = normalise_spacing(cspace)
    def set_chord_equal_distribution(self, numc: int):
        from pyvlm.tools import equal_spacing
        self.cspace = equal_spacing(numc)
    def set_chord_cosine_distribution(self, numc: int):
        from pyvlm.tools import full_cosine_spacing
        if numc > 1:
            spc = full_cosine_spacing(4*numc+2)
            self.cspace = spc[1::4]
            self.cspace[0] = spc[0]
            self.cspace[-1] = spc[-1]
            self.xspace = spc[2::4]
        else:
            self.cspace = [0.0, 1.0]
            self.xspace = [0.25]
    def set_chord_elliptical_distribution(self, numc: int):
        from pyvlm.tools import full_elliptical_spacing
        self.cspace = full_elliptical_spacing(numc)
    def mesh(self, lsid: int, lpid: int):
        from pygeom.geom3d import Point
        from numpy.matlib import empty
        nums = len(self.sects)
        self.shts = []
        for i in range(nums-1):
            a, b = i, i+1
            secta = self.sects[a]
            sectb = self.sects[b]
            self.shts.append(LatticeSheet(secta, sectb))
        self.strps = []
        for sht in self.shts:
            lsid = sht.mesh_strips(lsid)
            self.strps += sht.strps
        pnts = [strp.pnt1 for strp in self.strps]
        pnts.append(self.strps[-1].pnt2)
        crds = [strp.crd1 for strp in self.strps]
        crds.append(self.strps[-1].crd2)
        lenb = len(pnts)
        lenc = len(self.cspace)
        self.pnts = empty((lenb, lenc), dtype=Point)
        for i in range(lenb):
            minx = pnts[i].x
            y = pnts[i].y
            z = pnts[i].z
            c = crds[i]
            for j in range(lenc):
                cd = self.cspace[j]
                x = minx+cd*c
                self.pnts[i, j] = Point(x, y, z)
        self.pnls = empty((lenb-1, lenc-1), dtype=LatticePanel)
        for i, strp in enumerate(self.strps):
            for j in range(lenc-1):
                pnts = [
                    self.pnts[i, j],
                    self.pnts[i+1, j],
                    self.pnts[i, j+1],
                    self.pnts[i+1, j+1]
                ]
                cspc = (self.xspace[j]-self.cspace[j])/(self.cspace[j+1]-self.cspace[j])
                pnl = LatticePanel(lpid, pnts, cspc=cspc, bspc=strp.bspc)
                self.pnls[i, j] = pnl
                lpid += 1
                strp.add_panel(pnl)
        if self.mirror:
            numstrp = len(self.strps)
            hlfstrp = int(numstrp/2)
            for i in range(hlfstrp):
                strp = self.strps[numstrp-1-i]
                mstrp = self.strps[i]
                strp.set_mirror(mstrp)
        for sht in self.shts:
            sht.set_control_points_and_normals()
        return lsid, lpid
    def point_xyz(self):
        from numpy.matlib import zeros
        x = zeros(self.pnts.shape)
        y = zeros(self.pnts.shape)
        z = zeros(self.pnts.shape)
        for i in range(self.pnts.shape[0]):
            for j in range(self.pnts.shape[1]):
                x[i, j] = self.pnts[i, j].x
                y[i, j] = self.pnts[i, j].y
                z[i, j] = self.pnts[i, j].z
        return x, y, z
    def return_panels(self):
        pnls = []
        shp = self.pnls.shape
        for i in range(shp[0]):
            for j in range(shp[1]):
                pnls.append(self.pnls[i, j])
        return pnls
    def print_geom(self):
        print(f'Printing {self.name:s} geometry.')
        print('Sections')
        for i, sect in enumerate(self.sects):
            print(f'{i:d}\t{sect.pnt:.5f}\t{sect.chord:.5f}')
        print('Sheets')
        for i, sht in enumerate(self.shts):
            print(f'{i:d}')
            print(f'{sht.sect1.pnt:.5f}\t{sht.sect1.chord:.5f}')
            print(f'{sht.sect2.pnt:.5f}\t{sht.sect2.chord:.5f}')
    def __repr__(self):
        return '<LatticeSurface {:s}>'.format(self.name)

def latticesurface_from_json(surfdata: dict, display: bool=False):
    from .latticesection import latticesecttion_from_json
    name = surfdata['name']
    if 'mirror' in surfdata:
        mirror = surfdata['mirror']
    else:
        mirror = False
    if display: print(f'Loading Surface: {name:s}')
    sects = []
    for sectdata in surfdata['sections']:
        sect = latticesecttion_from_json(sectdata)
        sects.append(sect)
    surf = LatticeSurface(name, sects, mirror)
    if 'numc' in surfdata and 'cspace' in surfdata:
        numc = surfdata['numc']
        cspace = surfdata['cspace']
        if cspace == 'equal':
            surf.set_chord_equal_distribution(numc)
        elif cspace == 'cosine':
            surf.set_chord_cosine_distribution(numc)
        elif cspace == 'elliptical':
            surf.set_chord_elliptical_distribution(numc)
    return surf



from inspect import getfile, currentframe
from os.path import abspath, join, basename
from glob import glob

path = abspath(getfile(currentframe())).replace('__init__.py', '')

query = join(path, '*.json')
filepaths = glob(query)

filedict = {basename(filepath): filepath for filepath in filepaths}

def load_package_file(filename):
    from ..classes.latticesystem import latticesystem_from_json
    filepath = filedict[filename]
    return latticesystem_from_json(filepath)



from math import pi, sqrt, acos, sin
from matplotlib.pyplot import figure

class Bell(object):
    lift = None
    span = None
    y = None
    _ym = None
    _s = None
    _sm = None
    _th = None
    _thm = None
    _speed = None
    _rho = None
    def __init__(self, span: float, y: list):
        self.span = span
        self.y = y
    def set_lift(self, lift: float):
        self.lift = lift
    def set_speed(self, speed: float):
        self._speed = speed
    def set_density(self, rho: float):
        self._rho = rho
    def set_ym(self, ym: list):
        self._ym = ym
    @property
    def s(self):
        if self._s is None:
            self._s = [2*yi/self.span for yi in self.y]
        return self._s
    @property
    def th(self):
        if self._th is None:
            self._th = [acos(si) for si in self.s]
        return self._th
    @property
    def speed(self):
        if self._speed is None:
            self._speed = 1.0
        return self._speed
    @property
    def rho(self):
        if self._rho is None:
            self._rho = 1.0
        return self._rho
    @property
    def drag(self):
        L = self.lift
        b = self.span
        V = self.speed
        rho = self.rho
        return 8*L**2/(3*pi*V**2*b**2*rho)
    @property
    def ym(self):
        if self._ym is None:
            self._ym = []
            for i in range(len(self.y)-1):
                a, b = i, i+1
                ya = self.y[a]
                yb = self.y[b]
                self._ym.append((ya+yb)/2)
        return self._ym
    @property
    def sm(self):
        if self._sm is None:
            self._sm = [2*yi/self.span for yi in self.ym]
        return self._sm
    @property
    def thm(self):
        if self._thm is None:
            self._thm = [acos(si) for si in self.sm]
        return self._thm
    def return_lift_forces(self):
        L = self.lift
        Li = []
        for i in range(len(self.th)-1):
            a, b = i, i+1
            tha = self.th[a]
            thb = self.th[b]
            Li.append(L*(3*tha - 3*thb - 2*sin(2*tha) + sin(4*tha)/4 + 2*sin(2*thb) - sin(4*thb)/4)/(3*pi))
        return Li
    def return_delta_y(self):
        dy = []
        for i in range(len(self.y)-1):
            a, b = i, i+1
            ya = self.y[a]
            yb = self.y[b]
            dy.append(yb-ya)
        return dy
    def lift_distribution(self):
        L = self.lift
        b = self.span
        return [16*L*sin(th)**3/(3*pi*b) for th in self.th]
    def drag_distribution(self):
        L = self.lift
        b = self.span
        V = self.speed
        rho = self.rho
        return [32*L**2*(sin(th) - sin(3*th))*sin(th)**2/(3*pi**2*V**2*b**3*rho) for th in self.th]
    def wash_distribution(self):
        L = self.lift
        b = self.span
        V = self.speed
        rho = self.rho
        return [4*L*(2*s**2-1)/(pi*V*b**2*rho) for s in self.s]
    def trefftz_wash_distribution(self):
        L = self.lift
        b = self.span
        V = self.speed
        rho = self.rho
        return [8*L*(2*s**2-1)/(pi*V*b**2*rho) for s in self.s]
    def shear_force_distribution(self):
        L = self.lift
        th = [thi if thi <= pi/2 else thi-pi for thi in self.th]
        return [L*(3*thi - 2*sin(2*thi) + sin(4*thi)/4)/(3*pi) for thi in th]
    def bending_moment(self):
        L = self.lift
        b = self.span
        th = [thi if thi <= pi/2 else thi-pi for thi in self.th]
        return [L*b*(3*thi**2 - sin(thi)**4 - 3*sin(thi)**2)*sin(thi)/(12*pi) for thi in th]
    def root_shear_force(self):
        L = self.lift
        return L/2
    def root_bending_moment(self):
        L = self.lift
        b = self.span
        return L*b*(-16 + 3*pi**2)/(48*pi)
    def return_phi(self):
        L = self.lift
        b = self.span
        V = self.speed
        rho = self.rho
        return [16*L*sin(th)**3/(3*pi*b)/rho/V for th in self.thm]
    def __str__(self):
        sf = self.root_shear_force()
        bm = self.root_bending_moment()
        outstr = ''
        outstr += f'Span = {self.span:g}\n'
        outstr += f'Lift = {self.lift:g}\n'
        outstr += f'Drag = {self.drag:g}\n'
        outstr += f'Speed = {self.speed:g}\n'
        outstr += f'Density = {self.rho:g}\n'
        outstr += f'Root Shear Force = {sf:g}\n'
        outstr += f'Root Bending Moment = {bm:g}\n'
        return outstr
    def plot_lift_distribution(self, ax=None):
        if ax is None:
            fig = figure(figsize=(12, 8))
            ax = fig.gca()
            ax.grid(True)
        ax.plot(self.y, self.lift_distribution(), label='Elliptical Theory')
        ax.legend()
        return ax
    def plot_drag_distribution(self, ax=None):
        if ax is None:
            fig = figure(figsize=(12, 8))
            ax = fig.gca()
            ax.grid(True)
        ax.plot(self.y, self.drag_distribution(), label='Elliptical Theory')
        ax.legend()
        return ax
    def plot_wash_distribution(self, ax=None):
        if ax is None:
            fig = figure(figsize=(12, 8))
            ax = fig.gca()
            ax.grid(True)
        ax.plot(self.y, self.wash_distribution(), label='Elliptical Theory')
        ax.legend()
        return ax
    def plot_trefftz_wash_distribution(self, ax=None):
        if ax is None:
            fig = figure(figsize=(12, 8))
            ax = fig.gca()
            ax.grid(True)
        ax.plot(self.y, self.trefftz_wash_distribution(), label='Elliptical Theory')
        ax.legend()
        return ax
